indexing
	description: "{
      A class resembling a hash table, but actually having a hash table
      and maintaining one or more inversions in synch
}"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2008, Amalasoft";
	date: "$Date: 2008/12/03 $";
	revision: "$Revision: 001$";
	howto: "{
    Create an instance of this class with a capacity and a number of
    inversions (alternate structures holding the same items).
    Add inversions as desired, up to the number defined at creation.
	 Once inversions are in place, use typical HASH_TABLE routines to
    add, delete and access items.
}"

class AEL_DS_COMPLEX_TABLE [G, H -> HASHABLE]

inherit
	AEL_DS_INDIRECT_TABLE [G, H]
		rename
			make as table_make,
			make_table as make_tables
		redefine
			make_tables, extend, remove, wipe_out
		end

create
	make

--|========================================================================
feature {NONE} -- Initialization
--|========================================================================

	make (c, ni: INTEGER) is
		do
			number_of_inversions := ni
			create inversions.make (ni)
			table_make (c)
		end

	--|--------------------------------------------------------------

	make_tables (n: INTEGER) is
		local
			i, lim: INTEGER
		do
			Precursor (n)
			lim := number_of_inversions
			from i := 1
			until i > lim
			loop
				inversions.extend (create {AEL_DS_LIST_AGENT_SORTED[G]}.make (Void))
				i := i + 1
			end
		end

--|========================================================================
feature -- Status
--|========================================================================

	number_of_inversions: INTEGER
			-- Number of inversion lists maintained for table

	--|--------------------------------------------------------------

	inversion_has (i: INTEGER; v: like item): BOOLEAN is
			-- Does the 'i_th' inversion list have the given item?
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i).has (v)
		end

	--|--------------------------------------------------------------

	inversion_lt_function (i: INTEGER): FUNCTION [ANY, TUPLE [G, G], BOOLEAN] is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i).lt_function
		end

--|========================================================================
feature -- Status setting
--|========================================================================

	set_inversion_lt_function (i: INTEGER; v: like inversion_lt_function) is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			inversions.i_th (i).set_lt_function (v)
		end

--|========================================================================
feature -- Access
--|========================================================================

	inversion_item (i: INTEGER): G is
			-- Item at cursor position in inversion list number 'i'
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i).item
		end

	--|--------------------------------------------------------------

	i_th_inversion (i: INTEGER): AEL_DS_LIST_AGENT_SORTED [G] is
			-- 'i'th elemtn in inversions list
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i)
		end

--|========================================================================
feature -- Cursor movement and state in inversion lists
--|========================================================================

	inversion_cursor (i: INTEGER): CURSOR is
			-- Cursor for given inversion list
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i).cursor
		end

	inversion_start (i: INTEGER) is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			inversions.i_th (i).start
		end

	inversion_finish (i: INTEGER) is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			inversions.i_th (i).start
		end

	inversion_forth (i: INTEGER) is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			inversions.i_th (i).forth
		end

	inversion_back (i: INTEGER) is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			inversions.i_th (i).back
		end

	inversion_after (i: INTEGER): BOOLEAN is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i).after
		end

	inversion_exhausted (i: INTEGER): BOOLEAN is
		require
			valid_index: i > 0 and i <= number_of_inversions
		do
			Result := inversions.i_th (i).exhausted
		end

--|========================================================================
feature -- Element change
--|========================================================================

	extend (v: G; k: H) is
		do
			table.extend (v, k)
			add_to_inversions (v, k)
		end

--|========================================================================
feature -- Element removal
--|========================================================================

	wipe_out is
			-- remove all items
		do
			Precursor
			wipe_out_inversions
		end

	--|--------------------------------------------------------------

	remove (k: H) is
			-- remove item with given key
		do
			remove_item_from_inversions (item(k), k)
			table.remove (k)
		end

--|========================================================================
feature {NONE} -- Implementation
--|========================================================================

	add_to_inversions (v: G; k: H) is
		require
			exists: v /= Void
			is_new: not table.has (k)
		do
			inversions.do_all (agent add_item_to_inversion (v,?))
		end

	--|--------------------------------------------------------------

	remove_item_from_inversions (v: G; k: H) is	
		require
			exists: v /= Void
			belongs: has (k)
		do
			inversions.do_all (agent remove_item_from_inversion (v,?))
		end

	--|--------------------------------------------------------------

	wipe_out_inversions is
		do
			inversions.do_all (agent wipe_out_inversion)
		end

--|========================================================================
feature {NONE} -- Agents for inversion iterators
--|========================================================================

	add_item_to_inversion (v: G; il: like i_th_inversion) is
			-- Add the given item 'v' to the inversion list 'il'
		require
			list_exists: il /= Void
		do
			il.extend (v)
		end

	--|--------------------------------------------------------------

	remove_item_from_inversion (v: G; il: like i_th_inversion) is
			-- Remove the given item 'v' from the inversion list 'il'
		require
			list_exists: il /= Void
		local
			oc: CURSOR
		do
			oc := il.cursor
			il.start
			il.search (v)
			if not il.exhausted then
				il.remove
			end
			if il.valid_cursor (oc) then
				il.go_to (oc)
			else
				il.start
			end
		end

	--|--------------------------------------------------------------

	wipe_out_inversion (il: like i_th_inversion) is
		require
			list_exists: il /= Void
		do
			il.wipe_out
		end

--|========================================================================
feature {AEL_DS_INDIRECT_TABLE} -- Structure
--|========================================================================

	inversions: ARRAYED_LIST [like i_th_inversion]

	--|--------------------------------------------------------------
invariant
	inversions_exist: inversions /= Void

end -- class AEL_DS_COMPLEX_TABLE


