indexing
	description: "{
      A doubley-linked list supporting sorting using client-defined agents.
      Items are not constrained to COMPARABLE
}"
	license: "Eiffel Forumn License v2 - See license.txt"
	source: "{
		Amalasoft Corporation
		273 Harwood Avenue
		Littleton, MA 01460 USA
		}"
	status: "Copyright 1995-2008, Amalasoft";
	date: "$Date: 2008/12/03 $";
	revision: "$Revision: 001$";
	howto: "{
    Create with a agent for comparing items
	 Use as you would a SORTED_TWO_WAY_LIST
}"

class AEL_DS_LIST_AGENT_SORTED [G]

inherit
	TWO_WAY_LIST [G]
		rename
			make as twl_make
		redefine
			has,
			prune_all, extend, new_chain
		end
	
create
	make

--|========================================================================
feature {NONE} -- Creation
--|========================================================================
	
	make (lt: like lt_function) is
			-- Create Current, setting the comparison function to 'lt'
		do
			set_lt_function (lt)
			twl_make
		end

 --|========================================================================
feature -- Access
 --|========================================================================

	has (v: G): BOOLEAN is
			-- Does structure include `v'?
 			-- (Reference or object equality, based on `object_comparison'.)
		local
			pos: CURSOR
		do
			if not is_empty then
				pos := cursor
				start
				search (v)
				Result := not after
				go_to (pos)
			end
		end

	--|--------------------------------------------------------------

	search_after (v: like item) is
			-- Go to first position with item greater
			-- than or equal to `v'.
		do
			from
				start
			until
				after or else less_or_equal (v, item)
			loop
				forth
			end
		ensure
			argument_less_than_item: (not after) implies less_or_equal (v, item)
		end

	--|--------------------------------------------------------------

	search_before (v: like item) is
			-- Go to last position with item less
			-- than or equal to `v'.
		do
			from
				finish
			until
				before or else greater_or_equal (v, item)
			loop
				back
			end
		ensure
			(not off) implies less_or_equal (item, v)
		end

 --|========================================================================
feature -- Element change
 --|========================================================================

	extend (v: like item) is
			-- Put `v' at proper position in list.
			-- The cursor ends up on the newly inserted
			-- item.
		do
			search_after (v)
			put_left (v)
			back
		ensure then
	 		remains_sorted: (old is_sorted) implies is_sorted
			item_inserted: item = v
		end

	--|--------------------------------------------------------------

	merge (other: LINEAR [G]) is
			-- Add all items from `other' at their proper positions.
		do
			from
				other.start
			until
				other.off
			loop
				extend (other.item)
				other.forth
			end
		ensure then
	 		remains_sorted: (old is_sorted) implies is_sorted
		end


	--|--------------------------------------------------------------

	prune_all (v: like item) is
			-- Remove all items identical to `v'.
			-- (Reference or object equality,
			-- based on `object_comparison'.)
			-- Leave cursor `off'.
		do
			from
				start
				search (v)
			until
				off or else less (v, item)
			loop
				remove
			end
			if not off then finish; forth end
		end


 --|========================================================================
feature -- Transformation
 --|========================================================================

	sort is
			-- Sort all items.
			-- Has O(`count' * log (`count')) complexity.
			--| Uses comb-sort (BYTE February '91)
		local
			no_change: BOOLEAN
			gap: INTEGER
			left_cell, cell: like first_element
			left_cell_item, cell_item: like item
		do
			if not is_empty then
				from
					gap := count * 10 // 13
				until
					gap = 0
				loop
					from
						no_change := False
						go_i_th (1 + gap)
					until
						no_change
					loop
						no_change := True
						from
							left_cell := first_element
							cell := active	-- position of first_element + gap
						until
							cell = Void
						loop
							left_cell_item := left_cell.item
							cell_item := cell.item
							if less (cell_item, left_cell_item) then
									-- Swap `left_cell_item' with `cell_item'
								no_change := False
								cell.put (left_cell_item)
								left_cell.put (cell_item)
							end
							left_cell := left_cell.right
							cell := cell.right
						end
					end
					gap := gap * 10 // 13
				end
			end
		end

 --|========================================================================
feature -- Status report
 --|========================================================================

	is_sorted: BOOLEAN is
			-- Is the structure sorted?
		local
			c: CURSOR
			prev: like item
		do
			Result := True
			if count > 1 then
				from
					c := cursor
					start
					check not off end
					prev := item
					forth
				until
					after or not Result
				loop
					Result := less_or_equal (prev, item)
					prev := item
					forth
				end
				go_to (c)
			end
		end

	--|--------------------------------------------------------------

	min: like item is
			-- Minimum item
		require
			not_empty: not is_empty
		do
			Result := first
		ensure
			min_is_first: Result = first
			-- smallest: For every item `it' in list, `Result' <= `it'
		end

	--|--------------------------------------------------------------

	max: like item is
			-- Maximum item
		require
			not_empty: not is_empty
		do
			Result := last
		ensure
			max_is_last: Result = last
			-- largest: For every item `it' in list, `it' <= `Result'
		end

	--|--------------------------------------------------------------

	median: like item is
			-- Median item
		require
			not_empty: not is_empty
		do
			Result := i_th ((count + 1) // 2)
		ensure
			median_definition: Result = i_th ((count + 1) // 2)
		end
	
--|========================================================================
feature -- Element comparison
--|========================================================================
	
	less (v1, v2: like item): BOOLEAN is
			-- Is v1 less than v2?
		require
			agent_exists: lt_function /= Void
		do
			lt_function.call ([v1, v2])
			Result := lt_function.last_result
		ensure then
			asymmetric: Result implies not less (v2, v1)
		end
	
	--|--------------------------------------------------------------

	less_or_equal (v1, v2: like item): BOOLEAN is
			-- Is v1 less than or equal to v2?
		do
			Result := not less (v2, v1)
		ensure then
			definition: Result = less (v1, v2) or v1.is_equal (v2)
		end

	--|--------------------------------------------------------------

	greater_or_equal (v1, v2: like item): BOOLEAN is
			-- Is v1 greater than or equal to v2?
		do
			Result := not less (v1, v2)
 		ensure then
			definition: Result = less_or_equal (v2, v1)
		end

--|========================================================================
feature -- Sorting agents
--|========================================================================
	
	lt_function: FUNCTION [ANY, TUPLE [like item, like item], BOOLEAN]

--|========================================================================
feature -- Sorting agent setting
--|========================================================================
	
	set_lt_function (v: like lt_function) is
		do
			lt_function := v
		end

 --|========================================================================
feature {AEL_DS_LIST_AGENT_SORTED} -- Implementation
 --|========================================================================

	new_chain: like Current is
			-- A newly created instance of the same type.
			-- This feature may be redefined in descendants so as to
			-- produce an adequately allocated and initialized object.
		do
			create Result.make (lt_function)
		end
	
end -- class AEL_DS_LIST_AGENT_SORTED
